BuffTheTree = LibStub("AceAddon-3.0"):NewAddon("BuffTheTree", "AceEvent-3.0", "AceConsole-3.0")
local grat = LibStub("LibGratuity-3.0")
local LSM = LibStub("LibSharedMedia-3.0")
local BuffTheTree = BuffTheTree
local defaults = { profile = {
	locked = false,
	blessings = {},
	incombat = false,
	x = 200,
	y = 200,
	anchor = "CENTER",
	parent = "UIParent",
	point = "BOTTOMLEFT",
	debug = false,
	config = false,
	showapplied = true,
	showinbgs = false,
	alwaystrue = true,
	imbue = false,
	growth = "DOWN",
	spacing = 0,
	scale = 1,
	font = "Friz Quadrata TT",
	fontsize = 10,
	fontoutline = {["MONOCHROME"] = false, ["THICKOUTLINE"] = false, ["OUTLINE"] = false},
	scalefont = true,
}}

function BuffTheTree:DebugMessage(msg, level)
	if not self.opt.debug then return end
	if not level then level = 2 end
	if level <= self.opt.debug then
		msg = "|cff7cfc00BuffTheTree debugging:|r "..msg;
		print(msg)
	end
end

function BuffTheTree:UpdateTimers()
	local self = BuffTheTree
	if not self.CVarCache then self.CVarCache = {} end
	local scale = GetCVar("uiScale")
	local usescale = GetCVar("useUiScale")
	local changed
	if scale ~= self.CVarCache.scale then
		self.CVarCache.scale = scale
		changed = true
	end
	if usescale ~= self.CVarCache.usescale then
		self.CVarCache.usescale = usescale
		changed = true
	end
	if changed then
		self:UpdateFrameScale()
	end
	for i, k in pairs(self:GetBuffFrames()) do
		if k.name then
			local expires = (select(7, UnitBuff("player", k.name)))
			if expires then
				local totalminutes = (expires-GetTime())/60
				local minutes = math.floor(totalminutes)
				local seconds = math.floor((totalminutes-minutes)*60)
				if seconds < 10 then seconds = "0"..seconds end
				if minutes < 10 then minutes = "0"..minutes end
				local text = minutes..":"..seconds
				if k.textcache ~= text then
					k.textcache = text
					k.text:SetText(text)
				end
			else
				self:UNIT_AURA()
			end
		end
		if k.slot then
			local duration
			if k.slot == GetInventorySlotInfo("MainHandSlot") then
				duration = (select(2, GetWeaponEnchantInfo()))
			elseif k.slot == GetInventorySlotInfo("SecondaryHandSlot") then
				duration = (select(5, GetWeaponEnchantInfo()))
			else
				return
			end
			local text = BuffTheTree:GetWeaponImbueDuration(duration)
			if text ~= k.textcache then
				k.textcache = text
				k.text:SetText(text)
			end
		end
	end
end

function BuffTheTree:GetEffectiveUIScale()
	local scale = GetCVar("uiScale")
	local effscale
	local usescale = GetCVar("useUiScale")
	local addonscale = self.opt.scale
	if usescale and usescale == "1" then
		effscale = 35*scale*addonscale
	else
		effscale = 31*addonscale
	end
	return effscale
end

function BuffTheTree:IsBuffEligible(class, num)
	if not num then self:DebugMessage("Eligibility check cancelled due to no required number being passed"); return false end
	if self.opt.config then self:DebugMessage("Config mode: Eligible"); return true end
	if self.opt.alwaystrue and not ((GetNumRaidMembers() > 0 and self.opt.showinbgs) or GetRealNumRaidMembers() > 0) then self:DebugMessage("Always true: Eligible"); return true end
	local numinraid = 0
	for i=1, GetNumRaidMembers() do
		if (select(2, UnitClass("raid"..i))) == class then
			numinraid = numinraid+1;
			if numinraid >= num then
				self:DebugMessage("There are more than "..num.." of "..class.." in the raid");
				return true
			end
		end
	end
	self:DebugMessage("There are not enough of "..class.." in the raid. (Found "..numinraid..", need "..num..")")
	return false
end

function BuffTheTree:GenerateOutlineString()
	local outline = self.opt.fontoutline
	local outlinestring
	for i, k in pairs(outline) do
		if k then
			if not outlinestring then
				outlinestring = i
			else
				outlinestring = outlinestring..", "..i
			end
		end
	end
	return outlinestring or ""
end

function BuffTheTree:GetSpecialClassFilter(filter)
	self:DebugMessage("Attempting to retrieve filter: "..filter);
	if filter == "mana" then
		local filtertable = {
			["WARRIOR"] = false,
			["ROGUE"] = false,
			["SHAMAN"] = true,
			["DRUID"] = true,
			["PALADIN"] = true,
			["HUNTER"] = true,
			["PRIEST"] = true,
			["WARLOCK"] = true,
			["DEATHKNIGHT"] = false,
			["MAGE"] = true,
		}
		self:DebugMessage("Successfully returned mana filter table");
		return filtertable
	end
	if filter == "all" then
		local filtertable = {
			["WARRIOR"] = true,
			["ROGUE"] = true,
			["SHAMAN"] = true,
			["DRUID"] = true,
			["PALADIN"] = true,
			["HUNTER"] = true,
			["PRIEST"] = true,
			["WARLOCK"] = true,
			["DEATHKNIGHT"] = true,
			["MAGE"] = true,
		}
		self:DebugMessage("Successfully returned all filter table");
		return filtertable
	end
end

function BuffTheTree:GetPoints(growth)
	if growth == "DOWN" then
		return "TOPLEFT", "BOTTOMLEFT"
	end
	if growth == "UP" then
		return "BOTTOMLEFT", "TOPLEFT"
	end
	if growth == "LEFT" then
		return "TOPRIGHT", "TOPLEFT"
	end
	if growth == "RIGHT" then
		return "TOPLEFT", "TOPRIGHT"
	end
end

function BuffTheTree:GetOffsets(growth, spacing)
	if growth == "DOWN" then
		return 0, (0-spacing)
	end
	if growth == "UP" then
		return 0, spacing
	end
	if growth == "RIGHT" then
		return spacing, 0
	end
	if growth == "LEFT" then
		return (0-spacing), 0
	end
end

function BuffTheTree:UpdateFrameLocations()
	local frames = self:GetBuffFrames()
	local anchor, relativeto = self:GetPoints(self.opt.growth)
	local offsetx, offsety = self:GetOffsets(self.opt.growth, self.opt.spacing)
	for i=2, #frames do
		frames[i]:ClearAllPoints()
		frames[i]:SetPoint(anchor, frames[(i-1)], relativeto, offsetx, offsety)
	end
end

function BuffTheTree:UpdateFrameScale()
	local scale = self.opt.scale
	local font = LSM:Fetch("font", self.opt.font)
	local scalefont = self.opt.scalefont
	local size = self.opt.fontsize
	local outline = self:GenerateOutlineString()
	for id, frame in pairs(self:GetBuffFrames()) do
		frame:SetWidth(100*scale)
		frame:SetHeight(35*scale)
		frame.icon:SetWidth(35*scale)
		frame.icon:SetHeight(35*scale)
		local backdrop = frame.icon:GetBackdrop()
		backdrop.tileSize = self:GetEffectiveUIScale()
		local r, g, b, a = frame.icon:GetBackdropColor()
		frame.icon:SetBackdrop(backdrop)
		frame.icon:SetBackdropColor(r, g, b, a)
		frame.text:SetWidth(65*scale)
		frame.text:SetHeight(65*scale)
		frame.text:SetFont(font, scalefont and (scale*size) or size, outline)
		frame.imbuebg:SetWidth(20*scale)
		frame.imbuebg:SetHeight(12*scale)
		frame.imbuebg:SetBackdrop({bgFile = "Interface\\Addons\\BuffTheTree\\blank", edgeFile = false, tile = true, tileSize = 10*scale, edgeSize = 0, insets = { left = 0, right = 0, top = 0, bottom = 0}})
		frame.imbuebg:SetBackdropColor(0, 0, 0)
		frame.imbue:SetWidth(35*scale)
		frame.imbue:SetHeight(35*scale)
		frame.imbue:SetFont("Fonts\\FRIZQT__.TTF", scale*10)
		frame.imbue:ClearAllPoints()
		frame.imbue:SetPoint("CENTER", frame.imbuebg, "CENTER", 0, 0.8*scale)
	end
end

function BuffTheTree:GetBuffFrames()
	local t = {}
	for i, k in pairs(self) do
		if i:find("buff%d+") then
			t[tonumber(string.match(i, "buff(%d+)"))] = k
		end
	end
	return t
end

function BuffTheTree:GetBlessingPrio(qblessing)
	for prio, blessing in pairs(self.opt.blessings) do
		if blessing == qblessing then
			self:DebugMessage("Blessing of "..qblessing.." is priority "..prio..". Returning that value...");
			return prio
		end
	end
	self:DebugMessage("Blessing of "..qblessing.." doesn't have a priority.");
end

function BuffTheTree:GetOptions()
	local opt = {
		display = {
			name = "Display settings",
			type = "group",
			order = 1,
			args = {
				locked = {
					name = "Locked",
					type = "toggle",
					get = function() return self.opt.locked end,
					set = function(info, value) self.opt.locked = value end,
					order = 1,
				},
				incombat = {
					name = "Show in combat",
					type = "toggle",
					get = function() return self.opt.incombat end,
					set = function(info, value) self.opt.incombat = value end,
					order = 2,
				},
				debug = {
					name = "Debug",
					type = "range",
					min = 0,
					max = 2,
					step = 1,
					get = function() return self.opt.debug or 0 end,
					set = function(info, value) self.opt.debug = value end,
					order = 3,
				},
				config = {
					name = "Config mode",
					type = "toggle",
					get = function() return self.opt.config end,
					set = function(info, value) self.opt.config = value self:UNIT_AURA() end,
					order = 4,
				},
				showinbg = {
					name = "Show in battlegrounds",
					type = "toggle",
					get = function() return self.opt.showinbgs end,
					set = function(info, value) self.opt.showinbgs = value self:UNIT_AURA() end,
					order = 5,
				},
				showapplied = {
					name = "Hide applied buffs",
					type = "toggle",
					get = function() return not self.opt.showapplied end,
					set = function(info, value) self.opt.showapplied = not value self:UNIT_AURA() end,
					order = 6,
				},
				alwaystrue = {
					name = "Always show applied",
					type = "toggle",
					get = function() return self.opt.alwaystrue end,
					set = function(info, value) self.opt.alwaystrue = value self:UNIT_AURA() end,
					disabled = function() return not self.opt.showapplied end,
					order = 7,
				},
				imbue = {
					name = "Show Weapon imbues",
					type = "toggle",
					get = function() return self.opt.imbue end,
					set = function(info, value) self.opt.imbue = value self:UNIT_AURA() end,
					order = 8,
				},
			},
		},
		frame = {
			name = "Frame settings",
			type = "group",
			order = 2,
			args = {
				scale = {
					name = "Scale",
					type = "range",
					order = 1,
					desc = "Set the scale of the buffs",
					get = function() return self.opt.scale or 1 end,
					set = function(info, value) self.opt.scale = value; self:UpdateFrameScale() end,
					min = 0.01,
					max = 2,
					step = 0.01,
					isPercent = true,
				},
				growth = {
					name = "Growth",
					type = "select",
					order = 2,
					desc = "Select the direction the frames should grow in",
					get = function() return self.opt.growth or "DOWN" end,
					set = function(info, value) self.opt.growth = value; self:UpdateFrameLocations() end,
					values = {["DOWN"] = "Down", ["UP"] = "Up", ["RIGHT"] = "Right", ["LEFT"] = "Left"},
				},
				spacing = {
					name = "Spacing",
					type = "range",
					order = 3,
					desc = "Set the buff spacing",
					get = function() return self.opt.spacing or 0 end,
					set = function(info, value) self.opt.spacing = value; self:UpdateFrameLocations() end,
					min = 0,
					max = 100,
					step = 0.2,
				},
			},
		},
		fontsettings = {
			name = "Font settings",
			type = "group",
			order = 3,
			args = {
				appearance = {
					name = "Appearance",
					type = "group",
					order = 1,
					dialogInline = true,
					args = {
						font = {
							name = "Font",
							type = "select",
							desc = "Modify the font used",
							get = function() return self.opt.font end,
							set = function(info, value) self.opt.font = value; self:UpdateFrameScale() end,
							values = LSM:HashTable("font"),
							dialogControl = "LSM30_Font",
							order = 1,
						},
						outline = {
							name = "Font outline",
							type = "multiselect",
							desc = "Modify the outline used by the font",
							get = function(info, key) return self.opt.fontoutline[key] end,
							set = function(info, key, value) self.opt.fontoutline[key] = value; self:UpdateFrameScale() end,
							values = {["OUTLINE"] = "Outline", ["THICKOUTLINE"] = "Thick outline", ["MONOCHROME"] = "Monochrome"},
							order = 2,
						},
					},
				},
				size = {
					name = "Size",
					type = "group",
					order = 2,
					dialogInline = true,
					args = {
						fontsize = {
							name = "Font size",
							type = "range",
							desc = "Set the size this font uses",
							get = function() return self.opt.fontsize end,
							set = function(info, value) self.opt.fontsize = value; self:UpdateFrameScale() end,
							min = 1,
							max = 32,
							step = 1,
						},
						scalefont = {
							name = "Scale font?",
							type = "toggle",
							desc = "Set whether or not the font size scales with the frame",
							get = function() return self.opt.scalefont end,
							set = function(info, value) self.opt.scalefont = value; self:UpdateFrameScale() end,
						},
					},
				},
			},
		},
		buffsettings = {
			name = "Buff settings",
			type = "group",
			order = 4,
			args = {
				blessings = {
					type = "group",
					name = "Blessings",
					dialogInline = true,
					order = 1,
					args = {
						first = {
							name = "First blessing",
							type = "select",
							get = function() return self.opt.blessings[1] or "none" end,
							set = function(info, value) if value == "none" then self.opt.blessings[1] = nil; self:UNIT_AURA() end for i=1, 4 do if i~=1 then if self.opt.blessings[i] == value then self.opt.blessings[i] = self.opt.blessings[1]; end end end self.opt.blessings[1] = value; self:UNIT_AURA() end,
							order = 1,
							values = {["kings"] = "Blessing of Kings", ["wisdom"] = "Blessing of Wisdom", ["might"] = "Blessing of Might", ["sanc"] = "Blessing of Sanctuary", ["none"] = "None"},
						},
						second = {
							name = "Second blessing",
							type = "select",
							get = function() return self.opt.blessings[2] or "none" end,
							set = function(info, value) if value == "none" then self.opt.blessings[2] = nil; self:UNIT_AURA() end for i=1, 4 do if i~=2 then if self.opt.blessings[i] == value then self.opt.blessings[i] = self.opt.blessings[2]; end end end self.opt.blessings[2] = value; self:UNIT_AURA() end,
							order = 2,
							values = {["kings"] = "Blessing of Kings", ["wisdom"] = "Blessing of Wisdom", ["might"] = "Blessing of Might", ["sanc"] = "Blessing of Sanctuary", ["none"] = "None"},
						},
						third = {
							name = "Third blessing",
							type = "select",
							get = function() return self.opt.blessings[3] or "none" end,
							set = function(info, value) if value == "none" then self.opt.blessings[3] = nil; self:UNIT_AURA() end for i=1, 4 do if i~=3 then if self.opt.blessings[i] == value then self.opt.blessings[i] = self.opt.blessings[3]; end end end self.opt.blessings[3] = value; self:UNIT_AURA() end,
							order = 3,
							values = {["kings"] = "Blessing of Kings", ["wisdom"] = "Blessing of Wisdom", ["might"] = "Blessing of Might", ["sanc"] = "Blessing of Sanctuary", ["none"] = "None"},
						},
						fourth = {
							name = "Fourth blessing",
							type = "select",
							get = function() return self.opt.blessings[4] or "none" end,
							set = function(info, value) if value == "none" then self.opt.blessings[4] = nil; self:UNIT_AURA() end for i=1, 4 do if i~=4 then if self.opt.blessings[i] == value then self.opt.blessings[i] = self.opt.blessings[4]; end end end self.opt.blessings[4] = value; self:UNIT_AURA() end,
							order = 4,
							values = {["kings"] = "Blessing of Kings", ["wisdom"] = "Blessing of Wisdom", ["might"] = "Blessing of Might", ["sanc"] = "Blessing of Sanctuary", ["none"] = "None"},
						},
					},
				},
			},
		},
	}
	return opt
end

function BuffTheTree:OnProfileEnable()
	self.opt = self.db.profile
end

function BuffTheTree:GetBuffTable()
	self:DebugMessage("Retrieving buff table");
	local buffs = {
		pwf = {
			buffinfo = {
				[48162] = {GetSpellInfo(48162), "PRIEST", 1},
				[48161] = {GetSpellInfo(48161), "PRIEST", 1},
			},
			classfilter = self:GetSpecialClassFilter("all"),
			texture = "interface/icons/spell_holy_wordfortitude",
		},
		motw = {
			buffinfo = {
				[48469] = {GetSpellInfo(48469), "DRUID", 1},
				[48470] = {GetSpellInfo(48470), "DRUID", 1},
			},
			classfilter = self:GetSpecialClassFilter("all"),
			texture = "interface/icons/spell_nature_giftofthewild",
		},
		kings = {
			buffinfo = {
				[20217] = {GetSpellInfo(20217), "PALADIN", self:GetBlessingPrio("kings")},
				[25898] = {GetSpellInfo(25898), "PALADIN", self:GetBlessingPrio("kings")},
			},
			classfilter = self:GetSpecialClassFilter("all"),
			texture = "interface/icons/spell_magic_magearmor",
		},
		wisdom = {
			buffinfo = {
				[48936] = {GetSpellInfo(48936), "PALADIN", self:GetBlessingPrio("wisdom")},
				[48938] = {GetSpellInfo(48938), "PALADIN", self:GetBlessingPrio("wisdom")},
			},
			classfilter = self:GetSpecialClassFilter("all"),
			texture = "interface/icons/spell_holy_sealofwisdom",
		},
		might = {
			buffinfo = {
				[48932] = {GetSpellInfo(48932), "PALADIN", self:GetBlessingPrio("might")},
				[48934] = {GetSpellInfo(48934), "PALADIN", self:GetBlessingPrio("might")},
			},
			classfilter = self:GetSpecialClassFilter("all"),
			texture = "interface/icons/spell_holy_fistofjustice",
		},
		sanctuary = {
			buffinfo = {
				[20911] = {GetSpellInfo(20911), "PALADIN", self:GetBlessingPrio("sanc")},
				[25899] = {GetSpellInfo(25899), "PALADIN", self:GetBlessingPrio("sanc")},
			},
			classfilter = self:GetSpecialClassFilter("all"),
			texture = "interface/icons/spell_holy_greaterblessingofsanctuary",
		},
		ai = {
			buffinfo = {
				[61316] = {GetSpellInfo(61316), "MAGE", 1},
				[43002] = {GetSpellInfo(43002), "MAGE", 1},
				[61024] = {GetSpellInfo(61024), "MAGE", 1},
				[42995] = {GetSpellInfo(42995), "MAGE", 1},
			},
			classfilter = self:GetSpecialClassFilter("mana"),
			texture = "interface/icons/spell_holy_arcaneintellect",
		},
		spirit = {
			buffinfo = {
				[48073] = {GetSpellInfo(48073), "PRIEST", 1},
				[48074] = {GetSpellInfo(48074), "PRIEST", 1},
			},
			classfilter = self:GetSpecialClassFilter("mana"),
			texture = "interface/icons/spell_holy_prayerofspirit",
		},
		sprot = {
			buffinfo = {
				[48170] = {GetSpellInfo(48170), "PRIEST", 1},
				[48169] = {GetSpellInfo(48169), "PRIEST", 1},
			},
			classfilter = self:GetSpecialClassFilter("all"),
			texture = "interface/icons/spell_holy_prayerofshadowprotection",
		},
		flask = {
			buffinfo = {
				[54212] = {GetSpellInfo(54212), (select(2, UnitClass("player"))), 1},
				[53755] = {GetSpellInfo(53755), (select(2, UnitClass("player"))), 1},
				[53758] = {GetSpellInfo(53758), (select(2, UnitClass("player"))), 1},
				[53760] = {GetSpellInfo(53760), (select(2, UnitClass("player"))), 1},
				[67019] = {GetSpellInfo(67019), (select(2, UnitClass("player"))), 1},
			},
			classfilter = self:GetSpecialClassFilter("all"),
			texture = "interface/icons/inv_alchemy_endlessflask_05",
		},
		bufffood = {
			buffinfo = {
				["food"] = {GetSpellInfo(57294), (select(2, UnitClass("player"))), 1},
			},
			classfilter = self:GetSpecialClassFilter("all"),
			texture = "interface/icons/spell_misc_food",
		},
	}
	self:DebugMessage("Returning buff table successful");
	return buffs
end

function BuffTheTree:StartMoving()
	local self = BuffTheTree
	if self.opt.locked then self:DebugMessage("Frames are locked, not moving"); return end
	self:DebugMessage("Moving frames now");
	self.buff1:StartMoving()
end

function BuffTheTree:StopMoving()
	local self = BuffTheTree
	self:DebugMessage("Frames aren't dragged anymore, save position")
	self.buff1:StopMovingOrSizing()
	self.opt.anchor, self.opt.parent, self.opt.point, self.opt.x, self.opt.y = self.buff1:GetPoint()
	if not self.opt.parent then self.opt.parent = UIParent end
	self.opt.parent = self.opt.parent:GetName()
	self:DebugMessage("Position saved");
end

function BuffTheTree:CreateFrame(i)
	self:DebugMessage("Attempting to create frame "..i);
	local scale = self.opt.scale
	local f = CreateFrame("Button", "BuffTheTreeBuff"..i, UIParent)
	f:SetWidth(100*scale)
	f:SetHeight(35*scale)
	if i == 1 then
		self:DebugMessage("Frame 1 is being anchored to UIParent. X: "..self.opt.x.." Y: "..self.opt.y, 1)
		f:SetScript("OnUpdate", BuffTheTree.UpdateTimers)
		if _G[self.opt.parent] then
			f:SetPoint(self.opt.anchor, _G["self.opt.parent"], self.opt.point, self.opt.x, self.opt.y)
		else
			f:SetPoint("BOTTOMLEFT", UIParent, "BOTTOMLEFT", ((UIParent:GetWidth()/2)-(50*scale)), ((UIParent:GetHeight()/2)-(17.5*scale)))
		end
		f:SetMovable(true)
	else
		local anchor, relativeto = self:GetPoints(self.opt.growth)
		local offsetx, offsety = self:GetOffsets(self.opt.growth, self.opt.spacing)
		f:SetPoint(anchor, self["buff"..(i-1)], relativeto, offsetx, offsety)
	end
	f:SetScript("OnEnter", function(self)
		if not (self.slot or self.name) then return end
		GameTooltip:SetOwner(self, "ANCHOR_CURSOR")
		GameTooltip:ClearAllPoints()
		--GameTooltip:SetPoint("TOPLEFT", self, "BOTTOMLEFT")
		if self.name then
			GameTooltip:ClearLines()
			local c = 1
			while true do
				if UnitBuff("player", c) == self.name then
					break;
				end
				if not UnitBuff("player", c) then
					c = nil;
					break;
				end
				c = c+1;
			end
			GameTooltip:SetUnitBuff("player", c)
		end
		if self.slot then
			GameTooltip:ClearLines()
			GameTooltip:SetInventoryItem("player", self.slot)
		end
		GameTooltip:Show()
	end)
	f:SetScript("OnLeave", function() GameTooltip:Hide(); end)
	f:RegisterForClicks("AnyUp")
	f:SetScript("OnMouseDown", BuffTheTree.StartMoving)
	f:SetScript("OnMouseUp", BuffTheTree.StopMoving)
	self["buff"..i] = f;
	self:DebugMessage("Trunk frame created");
	self:DebugMessage("Creating Icon");
	local f2 = CreateFrame("Frame", "BuffTheTreeBuff"..i.."Icon", f)
	f2:SetWidth(35*scale)
	f2:SetHeight(35*scale)
	f2:SetPoint("TOPLEFT", f, 0, 0)
	f.icon = f2;
	self:DebugMessage("Icon created");
	self:DebugMessage("Creating fontstring");
	local font = LSM:Fetch("font", self.opt.font)
	local scalefont = self.opt.scalefont
	local size = self.opt.fontsize
	local outline = self:GenerateOutlineString()
	local f3 = f2:CreateFontString(nil, "ARTWORK")
	f3:SetFont(font, scalefont and (scale*size) or size, outline)
	f3:SetHeight(35*scale)
	f3:SetWidth(65*scale)
	f3:ClearAllPoints()
	f3:SetPoint("RIGHT", f, "RIGHT", 0, 0)
	f3:SetJustifyH("LEFT");
	f.text = f3;
	self:DebugMessage("Fontstring created");
	self:DebugMessage("Creating fontstring background")
	local f4 = CreateFrame("Frame", nil, f2)
	f4:SetWidth(20*scale)
	f4:SetHeight(12*scale)
	f4:ClearAllPoints()
	f4:SetPoint("CENTER", f2, "CENTER", 0, 0)
	f4:SetBackdrop({bgFile = "Interface\\Addons\\BuffTheTree\\blank", edgeFile = false, tile = true, tileSize = 10*scale, edgeSize = 0, insets = { left = 0, right = 0, top = 0, bottom = 0}})
	f4:SetBackdropColor(0, 0, 0)	
	f.imbuebg = f4
	self:DebugMessage("Fontstring background created")
	self:DebugMessage("Creating secondary fontstring")
	local f5 = f4:CreateFontString(nil, "ARTWORK")
	f5:SetFont("Fonts\\FRIZQT__.TTF", scale*10)
	f5:SetTextColor(1, 1, 1)
	f5:SetHeight(35*scale)
	f5:SetWidth(35*scale)
	f5:ClearAllPoints()
	f5:SetPoint("CENTER", f4, "CENTER", 0, 0.8*scale)
	f5:SetJustifyH("CENTER")
	f.imbue = f5
	self:DebugMessage("Secondary fontstring created")
	self:DebugMessage("Done creating frame "..i, 1)
	if not InCombatLockdown() or self.opt.incombat then
		f:Show()
	end
	f2:Show()
	f3:Show()
	f4:Show()
	f5:Show()
end

function BuffTheTree:IsInRaid()
	if self.opt.showinbgs and (GetNumRaidMembers() > 0) then
		return true
	end
	if (GetRealNumRaidMembers() > 0) then
		return true
	end
	if self.opt.config then
		return true
	end
	if self.opt.alwaystrue then
		return true
	end
	return false
end

function BuffTheTree:ShouldShowApplied()
	if self.opt.showapplied then return true end
end

function BuffTheTree:ShouldShowMissing()
	if self.opt.showinbgs and (GetNumRaidMembers() > 0) then
		return true
	end
	if (GetRealNumRaidMembers() > 0) then
		return true
	end
	if self.opt.config then
		return true
	end
	return false
end

function BuffTheTree:ShouldOverrideRequired()
	if not self.opt.showapplied then return false end
	if self.opt.config then
		return true
	end
	if self.opt.alwaystrue then
		return true
	end
	return false
end

function BuffTheTree:ShouldShowWeaponImbues()
	if not self.opt.imbue then
		return false
	end
	local class = (select(2, UnitClass("player")))
	if class == "ROGUE" or class == "SHAMAN" then
		return true
	end
	return false
end

function BuffTheTree:GetWeaponImbueDuration(duration)
	local totalminutes = duration/60000
	local minutes = math.floor(totalminutes)
	local seconds = math.floor((totalminutes-minutes)*60)
	if seconds < 10 then seconds = "0"..seconds end
	if minutes < 10 then minutes = "0"..minutes end
	local text = minutes..":"..seconds
	return text
end

function BuffTheTree:GetShamanTexture(imbue)
	local name, texture
	local imbues = {58804, 51994, 58790, 10399, 58796}
	for _, k in pairs(imbues) do
		name, _, texture = GetSpellInfo(k)
		if string.match(name, imbue) then break; end
	end
	if name then
		return texture
	end
end

function BuffTheTree:GetWeaponImbueNameTexture(slot)
	grat:SetInventoryItem("player", slot)
	local texture, name
	_, _, name = grat:Find("^(.+) %(%d+ [^%)]+%)$")
	if not name then return end
	if (select(2, UnitClass("player"))) == "ROGUE" then
		texture = (select(10, GetItemInfo(name)))
		if not texture then return end
		return name, texture
	elseif (select(2, UnitClass("player"))) == "SHAMAN" then
		name = strmatch(name, "%S+")
		texture = self:GetShamanTexture(name)
		if name and texture then
			return name, texture
		end
	end
end

function BuffTheTree:GetMissingTexture(class, offhand)
	local tex
	if class == "ROGUE" then
		tex = "interface/icons/ability_poisons"
	elseif class == "SHAMAN" then
		tex = GetInventoryItemTexture("player", GetInventorySlotInfo(offhand and "SecondaryHandSlot" or "MainHandSlot"))
	end
	if tex then return tex end
end

function BuffTheTree:UNIT_AURA(event, unit)
	if unit ~= "player" and (event == "UNIT_AURA" or event == "UNIT_INVENTORY_CHANGED") then self:DebugMessage("Unit isn't player"); return end
	for c, k in pairs(self:GetBuffFrames()) do
		k:Hide()
	end
	if InCombatLockdown() and not self.opt.incombat then return end
	if not self:IsInRaid() then return end
	local buffs = self:GetBuffTable()
	local i = 1
	for _, data in pairs(buffs) do
		if data.classfilter[(select(2, UnitClass("player")))] then
			local found, elig
			for id, info in pairs(data.buffinfo) do
				if self:IsBuffEligible(info[2], info[3]) or self:ShouldOverrideRequired() then
					if self:IsBuffEligible(info[2], info[3]) then
						elig = true
					end
					local texture, expires
					--self:DebugMessage("Retrieving buff with name "..info[1])
					_, _, texture, _, _, _, expires = UnitBuff("player", info[1])
					if texture and expires then
						if not self["buff"..i] then
							self:CreateFrame(i)
						end
						local f = self["buff"..i]
						self:DebugMessage(info[1].." found.");
						self:DebugMessage("Using "..f:GetName().." for "..info[1], 1)
						f.name = info[1]
						f.slot = nil
						self:DebugMessage("Texture is "..texture);
						f.icon:SetBackdrop({bgFile = texture, edgeFile = false, tile = true, tileSize = self:GetEffectiveUIScale(), edgeSize = 0, insets = { left=0, right=0, top=0, bottom=0}});
						f.icon:SetBackdropColor(1, 1, 1)
						local totalminutes = (expires-GetTime())/60
						local minutes = math.floor(totalminutes)
						local seconds = math.floor((totalminutes-minutes)*60)
						if seconds < 10 then seconds = "0"..seconds end
						if minutes < 10 then minutes = "0"..minutes end
						local text = minutes..":"..seconds
						self:DebugMessage("Text is "..text)
						f.text:SetText(text)
						f.text:SetTextColor(1, 1, 1)
						f.imbue:SetText(nil)
						f.imbuebg:Hide()
						if self:ShouldShowApplied() then
							f:Show()
							i = i+1;
						end
						found = true
						--self:DebugMessage(info[1].." set to displayed");
						break;
					end
				end
			end
			if (not found and elig) and self:ShouldShowMissing() then
				if not self["buff"..i] then
					self:CreateFrame(i);
				end
				local f = self["buff"..i]
				self:DebugMessage("Buff with texture "..data.texture.." not found.");
				self:DebugMessage("Using "..f:GetName().." for missing buff with texture "..data.texture, 1)
				f.name = nil
				f.slot = nil
				f.icon:SetBackdrop({bgFile = data.texture, edgeFile = false, tile = true, tileSize = self:GetEffectiveUIScale(), edgeSize = 0, insets = { left = 0, right = 0, top = 0, bottom = 0}});
				f.icon:SetBackdropColor(0.6, 0, 0)
				f.text:SetText("MISSING")
				f.text:SetTextColor(0.6, 0, 0)
				f.imbue:SetText(nil)
				f.imbuebg:Hide()
				f:Show()
				i=i+1
				--self:DebugMessage("Successfully set to MISSING");
			end
		end
	end
	if not self:ShouldShowWeaponImbues() then return end
	local hasMainHandEnchant, mainHandExpiration, mainHandCharges, hasOffHandEnchant, offHandExpiration, offHandCharges = GetWeaponEnchantInfo()
	if hasMainHandEnchant and mainHandCharges == 0 then
		local name, texture = self:GetWeaponImbueNameTexture(GetInventorySlotInfo("MainHandSlot"))
		local text = self:GetWeaponImbueDuration(mainHandExpiration)
		if not self["buff"..i] then
			self:CreateFrame(i)
		end
		local f = self["buff"..i]
		self:DebugMessage("Weapon imbue: "..name, 2)
		self:DebugMessage("Using "..f:GetName().." for "..name, 1)
		f.slot = GetInventorySlotInfo("MainHandSlot")
		f.name = nil
		f.icon:SetBackdrop({bgFile = texture, edgeFile = false, tile=true, tileSize = self:GetEffectiveUIScale(), edgeSize = 0, insets = { left=0, right=0, top=0, bottom=0}})
		f.icon:SetBackdropColor(1, 1, 1)
		f.text:SetText(text)
		f.text:SetTextColor(1, 1, 1)
		f.imbue:SetText("MH")
		f.imbuebg:Show()
		f:Show()
		i=i+1;
	else
		if not self["buff"..i] then
			self:CreateFrame(i)
		end
		local f = self["buff"..i]
		local texture = self:GetMissingTexture((select(2, UnitClass("player"))))
		self:DebugMessage("Using "..f:GetName().." for missing buff with texture "..texture, 1)
		if not self["buff"..i] then
			self:CreateFrame(i)
		end
		local f = self["buff"..i]
		f.name = nil
		f.slot = nil
		f.icon:SetBackdrop({bgFile = texture, edgeFile = false, tile = true, tileSize = self:GetEffectiveUIScale(), edgeSize = 0, insets = { left = 0, right = 0, top = 0, bottom = 0}})
		f.icon:SetBackdropColor(0.6, 0, 0)
		f.text:SetText("MISSING")
		f.text:SetTextColor(0.6, 0, 0)
		f.imbue:SetText("MH")
		f.imbuebg:Show()
		f:Show()
		i=i+1;
	end
	if hasOffHandEnchant and offHandCharges == 0 then
		local name, texture = self:GetWeaponImbueNameTexture(GetInventorySlotInfo("SecondaryHandSlot"))
		local text = self:GetWeaponImbueDuration(offHandExpiration)
		if not self["buff"..i] then
			self:CreateFrame(i)
		end
		local f = self["buff"..i]
		self:DebugMessage("Weapon imbue: "..name, 2)
		self:DebugMessage("Using "..f:GetName().." for "..name, 1)
		f.slot = GetInventorySlotInfo("SecondaryHandSlot")
		f.name = nil
		f.icon:SetBackdrop({bgFile = texture, edgeFile = false, tile=true, tileSize = self:GetEffectiveUIScale(), edgeSize = 0, insets = { left=0, right=0, top=0, bottom=0}})
		f.icon:SetBackdropColor(1, 1, 1)
		f.text:SetText(text)
		f.text:SetTextColor(1, 1, 1)
		f.imbue:SetText("OH")
		f.imbuebg:Show()
		f:Show()
		i=i+1;
	else
		if not self["buff"..i] then
			self:CreateFrame(i)
		end
		local f = self["buff"..i]
		local texture = self:GetMissingTexture((select(2, UnitClass("player"))), true)
		self:DebugMessage("Using "..f:GetName().." for missing buff with texture "..texture, 1)
		if not self["buff"..i] then
			self:CreateFrame(i)
		end
		local f = self["buff"..i]
		f.name = nil
		f.slot = nil
		f.icon:SetBackdrop({bgFile = texture, edgeFile = false, tile = true, tileSize = self:GetEffectiveUIScale(), edgeSize = 0, insets = { left = 0, right = 0, top = 0, bottom = 0}})
		f.icon:SetBackdropColor(0.6, 0, 0)
		f.text:SetText("MISSING")
		f.text:SetTextColor(0.6, 0, 0)
		f.imbue:SetText("OH")
		f.imbuebg:Show()
		f:Show()
		i=i+1;
	end
end

function BuffTheTree:OnInitialize()
	local openOptions = function()
		LibStub("AceConfigDialog-3.0"):Open("BuffTheTree")
	end
	self:RegisterChatCommand("btt", openOptions)
	self.db = LibStub("AceDB-3.0"):New("BuffTheTreeDB", defaults)
	self.db.RegisterCallback(self, "OnProfileChanged", "OnProfileEnable")
	self.db.RegisterCallback(self, "OnProfileCopied", "OnProfileEnable")
	self.db.RegisterCallback(self, "OnProfileReset", "OnProfileEnable")
	self.options = {type = 'group', args = {}}
	self.options.args = self:GetOptions()
	self.options.args.profiles = LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db)
	self.options.args.profiles.order = -1,
	LibStub("AceConfig-3.0"):RegisterOptionsTable("BuffTheTree", self.options)
	self.opt = self.db.profile
	self.enabled = false
	self.frames = {}
	self:CreateFrame(1)
	self:RegisterEvent("UNIT_AURA")
	self:RegisterEvent("PLAYER_ENTERING_WORLD", "UNIT_AURA")
	self:RegisterEvent("PLAYER_REGEN_DISABLED", "UNIT_AURA")
	self:RegisterEvent("PLAYER_REGEN_ENABLED", "UNIT_AURA")
	self:RegisterEvent("PARTY_MEMBERS_CHANGED", "UNIT_AURA")
	self:RegisterEvent("UNIT_INVENTORY_CHANGED", "UNIT_AURA")
end
